//  ICE Revision: $Id: FaFieldValueExpressionDriver.C,v a07f94cd351c 2012-10-03 14:36:06Z bgschaid $

#include "FaFieldValueExpressionDriver.H"
#include "FaFieldValuePluginFunction.H"

#include <Random.H>
#include <dimensionedVector.H>
#include "zeroGradientFaPatchFields.H"
#include "fixedValueFaPatchFields.H"

#include "addToRunTimeSelectionTable.H"

namespace Foam {

word FaFieldValueExpressionDriver::driverName_="internalFaField";

defineTypeNameAndDebug(FaFieldValueExpressionDriver, 0);

addNamedToRunTimeSelectionTable(CommonValueExpressionDriver, FaFieldValueExpressionDriver, dictionary, internalFaField);
addNamedToRunTimeSelectionTable(CommonValueExpressionDriver, FaFieldValueExpressionDriver, idName, internalFaField);

FaFieldValueExpressionDriver::FaFieldValueExpressionDriver (
    const word &id,
    const fvMesh &mesh
)
    : FaCommonValueExpressionDriver(
        false,
        true,
        false
    ),
      mesh_(faRegionMesh(mesh)),
      typ_("nothing"),
      isLogical_(false),
      isSurfaceField_(false),
      resultDimension_(0,0,0,0,0,0,0)
{
}

FaFieldValueExpressionDriver::FaFieldValueExpressionDriver (
    const fvMesh &mesh,
    bool cacheReadFields,
    bool searchInMemory,
    bool searchOnDisc
)
    : FaCommonValueExpressionDriver(
        cacheReadFields,
        searchInMemory,
        searchOnDisc
    ),
      mesh_(faRegionMesh(mesh)),
      typ_("nothing"),
      isLogical_(false),
      isSurfaceField_(false),
      resultDimension_(0,0,0,0,0,0,0)
{
}

FaFieldValueExpressionDriver::FaFieldValueExpressionDriver (
    const dictionary &dict,
    const fvMesh &mesh
)
    : FaCommonValueExpressionDriver(dict),
      mesh_(
          faRegionMesh(
              regionMesh(
                  dict,
                  mesh,
                  searchOnDisc()
              ))),
      typ_("nothing"),
      isLogical_(false),
      isSurfaceField_(false),
      resultDimension_(0,0,0,0,0,0,0)
{
    if(dict.found("dimensions")) {
        resultDimension_.reset(dimensionSet(dict.lookup("dimensions")));
    }
}

FaFieldValueExpressionDriver::~FaFieldValueExpressionDriver ()
{
}

void FaFieldValueExpressionDriver::parseInternal (int startToken)
{
    parserFaField::FaFieldValueExpressionParser parser (
        scanner_,
        *this,
        startToken,
        0
    );
    parser.set_debug_level (traceParsing());
    parser.parse ();
}

areaScalarField *FaFieldValueExpressionDriver::makeModuloField(
    const areaScalarField &a,
    const areaScalarField &b)
{
    areaScalarField *result_=makeConstantField<areaScalarField>(0.);

    forAll(*result_,cellI) {
        scalar val=fmod(a[cellI],b[cellI]);

        if(fabs(val)>(b[cellI]/2)) {
            if(val>0) {
                val-=b[cellI];
            } else {
                val=b[cellI];
            }
        }

        (*result_)[cellI]=val;
    }

    return result_;
}

areaScalarField *FaFieldValueExpressionDriver::makeRandomField(label seed)
{
    areaScalarField *f=makeConstantField<areaScalarField>(0.);

    autoPtr<scalarField> rField(CommonValueExpressionDriver::makeRandomField(seed));
    f->internalField()=rField();

    return f;
}

areaScalarField *FaFieldValueExpressionDriver::makeCellIdField()
{
    areaScalarField *f=makeConstantField<areaScalarField>(0.);

    forAll(*f,cellI) {
        (*f)[cellI]=scalar(cellI);
    }

    return f;
}

areaScalarField *FaFieldValueExpressionDriver::makeGaussRandomField(label seed)
{
    areaScalarField *f=makeConstantField<areaScalarField>(0.);

    autoPtr<scalarField> rField(CommonValueExpressionDriver::makeGaussRandomField(seed));
    f->internalField()=rField();

    return f;
}

areaVectorField *FaFieldValueExpressionDriver::makePositionField()
{
    dimensionSet nullDim(0,0,0,0,0);
    areaVectorField *f=new areaVectorField(
        IOobject
        (
            "pos",
            time(),
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        aMesh(),
        vector(0,0,0)
    );
    f->dimensions().reset(mesh_.areaCentres().dimensions());
    *f=mesh_.areaCentres();
    f->dimensions().reset(nullDim);
    return f;
}

edgeVectorField *FaFieldValueExpressionDriver::makeEdgePositionField()
{
    dimensionSet nullDim(0,0,0,0,0);
    edgeVectorField *f=new edgeVectorField(
        IOobject
        (
            "fpos",
            time(),
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        aMesh(),
        vector(0,0,0)
    );
    f->dimensions().reset(mesh_.edgeCentres().dimensions());
    *f=mesh_.edgeCentres();
    f->dimensions().reset(nullDim);
    return f;
}

edgeVectorField *FaFieldValueExpressionDriver::makeEdgeProjectionField()
{

    dimensionSet nullDim(0,0,0,0,0);
    edgeVectorField *f=new edgeVectorField(
        IOobject
        (
            "fproj",
            time(),
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        aMesh(),
        vector(0,0,0)
    );
    f->dimensions().reset(mesh_.edgeCentres().dimensions());

    vector fmin(0,0,0);
    vector fmax(0,0,0);

    forAll(*f,faceI)
    {
        const edge &fProp = mesh_.edges()[faceI];
        fmin = mesh_.points()[fProp[0]];
        fmax = fmin;
        forAll(fProp,pointI)
        {
            forAll(mesh_.points()[0],compI)
            {
                if(
                    mesh_.points()[fProp[pointI]].component(compI)
                    <
                    fmin.component(compI)
                ) {
                    fmin.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                }
                if(
                    mesh_.points()[fProp[pointI]].component(compI)
                    >
                    fmax.component(compI)
                ) {
                    fmax.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                }
            }
        }
        (*f)[faceI] = fmax - fmin;
    }
    forAll(mesh_.boundary(),patchI)
    {
        labelList cNumbers = mesh_.boundary()[patchI].edgeFaces();
        faePatchVectorField & fFace = f->boundaryField()[patchI];

        forAll(fFace,faceI)
        {
            const face & cProp(mesh_.faces()[cNumbers[faceI]]);

            forAll(cProp,cJ)
            {
                const label patchID = mesh_.boundary().whichPatch(cProp[cJ]);

                if (patchID == patchI)
                {
                    const edge & fProp = mesh_.edges()[cProp[cJ]];

                    fmin = mesh_.points()[fProp[0]];
                    fmax = fmin;

                    forAll(fProp,pointI)
                    {
                        forAll(mesh_.points()[0],compI)
                        {
                            if(
                                mesh_.points()[fProp[pointI]].component(compI)
                                <
                                fmin.component(compI)
                            ) {
                                fmin.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                            }
                            if(
                                mesh_.points()[fProp[pointI]].component(compI)
                                >
                                fmax.component(compI)
                            ) {
                                fmax.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                            }
                        }
                    }
                    fFace[faceI] = fmax - fmin;
                }
            }
        }
    }

    f->dimensions().reset(nullDim);
    return f;
}

edgeVectorField *FaFieldValueExpressionDriver::makeEdgeField()
{
    dimensionSet nullDim(0,0,0,0,0);
    edgeVectorField *f=new edgeVectorField(
        IOobject
        (
            "face",
            time(),
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        aMesh(),
        vector(0,0,0)
    );
    f->dimensions().reset(mesh_.Le().dimensions());
    *f=mesh_.Le();
    f->dimensions().reset(nullDim);
    return f;
}

edgeScalarField *FaFieldValueExpressionDriver::makeLengthField()
{
    dimensionSet nullDim(0,0,0,0,0);
    edgeScalarField *f=new edgeScalarField(
        IOobject
        (
            "face",
            time(),
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        aMesh(),
        0.
    );
    f->dimensions().reset(mesh_.magLe().dimensions());
    *f=mesh_.magLe();
    f->dimensions().reset(nullDim);
    return f;
}

areaScalarField *FaFieldValueExpressionDriver::makeAreaField()
{
    areaScalarField *f=new areaScalarField(
        IOobject
        (
            "area",
            time(),
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        aMesh(),
        0.
    );
    const scalarField &V=mesh_.S();

    forAll(*f,cellI) {
        (*f)[cellI]=V[cellI];
    }

    return f;
}

areaScalarField *FaFieldValueExpressionDriver::makeRDistanceField(const areaVectorField& r)
{
    dimensionSet nullDim(0,0,0,0,0);
    areaScalarField *f=new areaScalarField(
        IOobject
        (
            "rdist",
            time(),
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        aMesh(),
        0.
    );

    forAll(*f,cellI) {
        (*f)[cellI]=mag(mesh_.areaCentres()[cellI] - r[cellI]);
    }

    return f;
}

areaVectorField *FaFieldValueExpressionDriver::makeVectorField
(
    areaScalarField *x,
    areaScalarField *y,
    areaScalarField *z
) {
    areaVectorField *f=makeConstantField<areaVectorField>(vector(0,0,0));

    forAll(*f,cellI) {
        (*f)[cellI]=vector((*x)[cellI],(*y)[cellI],(*z)[cellI]);
    }

    return f;
}

edgeVectorField *FaFieldValueExpressionDriver::makeEdgeVectorField
(
    edgeScalarField *x,
    edgeScalarField *y,
    edgeScalarField *z
)
{
    edgeVectorField *f=makeConstantField<edgeVectorField>(vector(0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=vector((*x)[faceI],(*y)[faceI],(*z)[faceI]);
    }

    return f;
}

areaTensorField *FaFieldValueExpressionDriver::makeTensorField
(
    areaScalarField *xx,areaScalarField *xy,areaScalarField *xz,
    areaScalarField *yx,areaScalarField *yy,areaScalarField *yz,
    areaScalarField *zx,areaScalarField *zy,areaScalarField *zz
) {
    areaTensorField *f=makeConstantField<areaTensorField>(tensor(0,0,0,0,0,0,0,0,0));

    forAll(*f,cellI) {
        (*f)[cellI]=tensor(
            (*xx)[cellI],(*xy)[cellI],(*xz)[cellI],
            (*yx)[cellI],(*yy)[cellI],(*yz)[cellI],
            (*zx)[cellI],(*zy)[cellI],(*zz)[cellI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

areaSymmTensorField *FaFieldValueExpressionDriver::makeSymmTensorField
(
    areaScalarField *xx,areaScalarField *xy,areaScalarField *xz,
    areaScalarField *yy,areaScalarField *yz,
    areaScalarField *zz
) {
    areaSymmTensorField *f=makeConstantField<areaSymmTensorField>(symmTensor(0,0,0,0,0,0));

    forAll(*f,cellI) {
        (*f)[cellI]=symmTensor(
            (*xx)[cellI],(*xy)[cellI],(*xz)[cellI],
            (*yy)[cellI],(*yz)[cellI],
            (*zz)[cellI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

areaSphericalTensorField *FaFieldValueExpressionDriver::makeSphericalTensorField
(
    areaScalarField *xx
) {
    areaSphericalTensorField *f=makeConstantField<areaSphericalTensorField>(sphericalTensor(0));

    forAll(*f,cellI) {
        (*f)[cellI]=sphericalTensor(
            (*xx)[cellI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

edgeTensorField *FaFieldValueExpressionDriver::makeEdgeTensorField
(
    edgeScalarField *xx,edgeScalarField *xy,edgeScalarField *xz,
    edgeScalarField *yx,edgeScalarField *yy,edgeScalarField *yz,
    edgeScalarField *zx,edgeScalarField *zy,edgeScalarField *zz
) {
    edgeTensorField *f=makeConstantField<edgeTensorField>(tensor(0,0,0,0,0,0,0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=tensor(
            (*xx)[faceI],(*xy)[faceI],(*xz)[faceI],
            (*yx)[faceI],(*yy)[faceI],(*yz)[faceI],
            (*zx)[faceI],(*zy)[faceI],(*zz)[faceI]
        );
    }

    return f;
}

edgeSymmTensorField *FaFieldValueExpressionDriver::makeEdgeSymmTensorField
(
    edgeScalarField *xx,edgeScalarField *xy,edgeScalarField *xz,
    edgeScalarField *yy,edgeScalarField *yz,
    edgeScalarField *zz
) {
    edgeSymmTensorField *f=makeConstantField<edgeSymmTensorField>(symmTensor(0,0,0,0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=symmTensor(
            (*xx)[faceI],(*xy)[faceI],(*xz)[faceI],
            (*yy)[faceI],(*yz)[faceI],
            (*zz)[faceI]
        );
    }

    return f;
}

edgeSphericalTensorField *FaFieldValueExpressionDriver::makeEdgeSphericalTensorField
(
    edgeScalarField *xx
) {
    edgeSphericalTensorField *f=makeConstantField<edgeSphericalTensorField>(sphericalTensor(0));

    forAll(*f,faceI) {
        (*f)[faceI]=sphericalTensor(
            (*xx)[faceI]
        );
    }

    return f;
}

template<>
FaFieldValueExpressionDriver::SymbolTable<FaFieldValueExpressionDriver>::SymbolTable()
:
StartupSymbols()
{
    // default value
    insert("",parserFaField::FaFieldValueExpressionParser::token::START_DEFAULT);

    insert(
        "areaScalarField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_SCALAR_COMMA
    );
    insert(
        "areaScalarField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_SCALAR_CLOSE
    );
    insert(
        "areaVectorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_VECTOR_COMMA
    );
    insert(
        "areaVectorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_VECTOR_CLOSE
    );
    insert(
        "areaTensorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_TENSOR_COMMA
    );
    insert(
        "areaTensorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_TENSOR_CLOSE
    );
    insert(
        "areaSymmTensorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_YTENSOR_COMMA
    );
    insert(
        "areaSymmTensorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_YTENSOR_CLOSE
    );
    insert(
        "areaSphericalTensorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_HTENSOR_COMMA
    );
    insert(
        "areaSphericalTensorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_HTENSOR_CLOSE
    );
    insert(
        "areaLogicalField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_LOGICAL_COMMA
    );
    insert(
        "areaLogicalField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_AREA_LOGICAL_CLOSE
    );

    insert(
        "edgeScalarField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_SCALAR_COMMA
    );
    insert(
        "edgeScalarField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_SCALAR_CLOSE
    );
    insert(
        "edgeVectorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_VECTOR_COMMA
    );
    insert(
        "edgeVectorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_VECTOR_CLOSE
    );
    insert(
        "edgeTensorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_TENSOR_COMMA
    );
    insert(
        "edgeTensorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_TENSOR_CLOSE
    );
    insert(
        "edgeSymmTensorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_YTENSOR_COMMA
    );
    insert(
        "edgeSymmTensorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_YTENSOR_CLOSE
    );
    insert(
        "edgeSphericalTensorField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_HTENSOR_COMMA
    );
    insert(
        "edgeSphericalTensorField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_HTENSOR_CLOSE
    );
    insert(
        "edgeLogicalField_SC",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_LOGICAL_COMMA
    );
    insert(
        "edgeLogicalField_CL",
        parserFaField::FaFieldValueExpressionParser::token::START_EDGE_LOGICAL_CLOSE
    );

    insert(
        "CL",
        parserFaField::FaFieldValueExpressionParser::token::START_CLOSE_ONLY
    );
    insert(
        "SC",
        parserFaField::FaFieldValueExpressionParser::token::START_COMMA_ONLY
    );
}

const FaFieldValueExpressionDriver::SymbolTable<FaFieldValueExpressionDriver> &FaFieldValueExpressionDriver::symbolTable()
{
    static SymbolTable<FaFieldValueExpressionDriver> actualTable;

    return actualTable;
}

int FaFieldValueExpressionDriver::startupSymbol(const word &name) {
    return symbolTable()[name];
}


autoPtr<CommonPluginFunction> FaFieldValueExpressionDriver::newPluginFunction(
    const word &name
) {
    return autoPtr<CommonPluginFunction>(
        FaFieldValuePluginFunction::New(
            *this,
            name
        ).ptr()
    );
}

bool FaFieldValueExpressionDriver::existsPluginFunction(
    const word &name
) {
    return FaFieldValuePluginFunction::exists(
        *this,
        name
    );
}

label FaFieldValueExpressionDriver::size() const
{
    return mesh_.nFaces();
}

label FaFieldValueExpressionDriver::pointSize() const
{
    return mesh_.nPoints();
}

const faMesh &FaFieldValueExpressionDriver::aMesh() const
{
    return mesh_;
}

const fvMesh &FaFieldValueExpressionDriver::mesh() const
{
    return dynamic_cast<const fvMesh&>(mesh_.thisDb());
}

} // end namespace
